/*
 * RoomsTerrainGeneratorPlugin.java 11 march 2021
 *
 * Sweet Home 3D, Copyright (c) 2021 Emmanuel PUYBARET / eTeks <info@eteks.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.eteks.sweethome3d.plugin.terraingenerator;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.Graphics;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsEnvironment;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.KeyboardFocusManager;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.geom.Area;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.io.File;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.lang.reflect.Field;
import java.net.URL;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.Format;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.concurrent.Callable;

import javax.jnlp.BasicService;
import javax.jnlp.ServiceManager;
import javax.jnlp.UnavailableServiceException;
import javax.media.j3d.AmbientLight;
import javax.media.j3d.Appearance;
import javax.media.j3d.Background;
import javax.media.j3d.BoundingBox;
import javax.media.j3d.BoundingSphere;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.Canvas3D;
import javax.media.j3d.ColoringAttributes;
import javax.media.j3d.DirectionalLight;
import javax.media.j3d.Geometry;
import javax.media.j3d.GeometryArray;
import javax.media.j3d.GraphicsConfigTemplate3D;
import javax.media.j3d.Group;
import javax.media.j3d.Light;
import javax.media.j3d.LineAttributes;
import javax.media.j3d.Node;
import javax.media.j3d.PolygonAttributes;
import javax.media.j3d.RenderingAttributes;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Texture;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.media.j3d.TransparencyAttributes;
import javax.media.j3d.View;
import javax.swing.AbstractAction;
import javax.swing.ActionMap;
import javax.swing.BorderFactory;
import javax.swing.InputMap;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JEditorPane;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.KeyStroke;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingUtilities;
import javax.swing.border.EtchedBorder;
import javax.swing.event.AncestorEvent;
import javax.swing.event.AncestorListener;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.event.MouseInputAdapter;
import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.UndoableEditSupport;
import javax.vecmath.Color3f;
import javax.vecmath.Point3d;
import javax.vecmath.Point3f;
import javax.vecmath.TexCoord2f;
import javax.vecmath.Vector3d;
import javax.vecmath.Vector3f;

import com.eteks.sweethome3d.j3d.Component3DManager;
import com.eteks.sweethome3d.j3d.ModelManager;
import com.eteks.sweethome3d.j3d.OBJWriter;
import com.eteks.sweethome3d.j3d.Object3DBranch;
import com.eteks.sweethome3d.j3d.Object3DBranchFactory;
import com.eteks.sweethome3d.j3d.TextureManager;
import com.eteks.sweethome3d.model.CatalogPieceOfFurniture;
import com.eteks.sweethome3d.model.Content;
import com.eteks.sweethome3d.model.Home;
import com.eteks.sweethome3d.model.HomeEnvironment;
import com.eteks.sweethome3d.model.HomeFurnitureGroup;
import com.eteks.sweethome3d.model.HomePieceOfFurniture;
import com.eteks.sweethome3d.model.HomeTexture;
import com.eteks.sweethome3d.model.InterruptedRecorderException;
import com.eteks.sweethome3d.model.Level;
import com.eteks.sweethome3d.model.RecorderException;
import com.eteks.sweethome3d.model.Room;
import com.eteks.sweethome3d.model.Selectable;
import com.eteks.sweethome3d.model.SelectionEvent;
import com.eteks.sweethome3d.model.SelectionListener;
import com.eteks.sweethome3d.model.UserPreferences;
import com.eteks.sweethome3d.model.Wall;
import com.eteks.sweethome3d.plugin.Plugin;
import com.eteks.sweethome3d.plugin.PluginAction;
import com.eteks.sweethome3d.swing.AutoCommitSpinner;
import com.eteks.sweethome3d.swing.SwingTools;
import com.eteks.sweethome3d.swing.SwingViewFactory;
import com.eteks.sweethome3d.tools.OperatingSystem;
import com.eteks.sweethome3d.tools.ResourceURLContent;
import com.eteks.sweethome3d.tools.TemporaryURLContent;
import com.eteks.sweethome3d.viewcontroller.HomeView;
import com.eteks.sweethome3d.viewcontroller.ThreadedTaskController;
import com.sun.j3d.exp.swing.JCanvas3D;
import com.sun.j3d.utils.geometry.Box;
import com.sun.j3d.utils.geometry.Cylinder;
import com.sun.j3d.utils.geometry.GeometryInfo;
import com.sun.j3d.utils.geometry.NormalGenerator;
import com.sun.j3d.utils.geometry.Sphere;
import com.sun.j3d.utils.picking.PickCanvas;
import com.sun.j3d.utils.picking.PickIntersection;
import com.sun.j3d.utils.picking.PickResult;
import com.sun.j3d.utils.universe.SimpleUniverse;
import com.sun.j3d.utils.universe.Viewer;
import com.sun.j3d.utils.universe.ViewingPlatform;

/**
 * A plug-in that creates a 3D terrain from elevations at different points of a room.
 * @author Emmanuel Puybaret
 */
public class RoomsTerrainGeneratorPlugin extends Plugin {
  private JPanel                  terrainPanel;
  private TerrainPreviewComponent terrainPreview;

  @Override
  public PluginAction [] getActions() {
    return new PluginAction [] {new PluginAction(
            "com.eteks.sweethome3d.plugin.terraingenerator.ApplicationPlugin",
            "GENERATE_TERRAIN", getPluginClassLoader(), false) {
        {
          // Add a listener to enable action only when selection isn't empty
          getHome().addSelectionListener(new SelectionListener() {
              public void selectionChanged(SelectionEvent ev) {
                setEnabled(areRoomsAtSameElevationSelected());
              }
            });
        }

        @Override
        public void execute() {
          if (areRoomsAtSameElevationSelected()) {
            generateTerrain();
          }
        }
      }
    };
  }

  private boolean areRoomsAtSameElevationSelected() {
    Home home = getHome();
    List<Selectable> selectedItems = home.getSelectedItems();
    Float elevation = null;
    if (selectedItems.size() > 0) {
      for (Selectable item : selectedItems) {
        if (!(item instanceof Room)) {
          return false;
        } else {
          // Check all rooms are at the same elevation
          Level level = ((Room)item).getLevel();
          float roomElevation = level != null ? level.getElevation() : 0;
          if (elevation == null) {
            elevation = roomElevation;
          } else if (elevation != roomElevation) {
            return false;
          }
        }
      }
      return true;
    } else {
      return false;
    }
  }

  private void generateTerrain() {
    final ResourceBundle resource = ResourceBundle.getBundle(
        "com.eteks.sweethome3d.plugin.terraingenerator.ApplicationPlugin",
        Locale.getDefault(), getPluginClassLoader());
    final HomeView homeView = getHomeController().getView();

    try {
      // Ignore plug-in in protected Java Web Start environment
      ServiceManager.lookup("javax.jnlp.FileSaveService");
      // Use an uneditable editor pane to let user select text in dialog
      JEditorPane messagePane = new JEditorPane("text/html",
          resource.getString("terrainGeneratorJavaWebStartInfo.message"));
      messagePane.setOpaque(false);
      messagePane.setEditable(false);
      try {
        // Lookup the javax.jnlp.BasicService object
        final BasicService service = (BasicService)ServiceManager.lookup("javax.jnlp.BasicService");
        // If basic service supports  web browser
        if (service.isWebBrowserSupported()) {
          // Add a listener that displays hyperlinks content in browser
          messagePane.addHyperlinkListener(new HyperlinkListener() {
            public void hyperlinkUpdate(HyperlinkEvent ev) {
              if (ev.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
                service.showDocument(ev.getURL());
              }
            }
          });
        }
      } catch (UnavailableServiceException ex) {
        // Too bad : service is unavailable
      }

      String title = resource.getString("terrainGeneratorJavaWebStartInfo.title");
      JOptionPane.showMessageDialog((JComponent)homeView, messagePane, title, JOptionPane.WARNING_MESSAGE);
      return;
    } catch (UnavailableServiceException ex) {
    }

    List<Selectable> selectedItems = getHome().getSelectedItems();
    // Get selected rooms in the order they were added in home
    final List<Room> selectedRooms = new ArrayList<Room>();
    for (Room room : getHome().getRooms()) {
      if (selectedItems.contains(room)) {
        selectedRooms.add(room);
      }
     }
    final List<Room> editedRooms = new ArrayList<Room>(selectedRooms.size());
    for (Room room : selectedRooms) {
      editedRooms.add(room.clone());
    }
    List<Selectable> homeItems = getHome().getSelectableViewableItems();
    homeItems.removeAll(selectedRooms);
    boolean otherVisibleHomeItems = false;
    for (Selectable item : homeItems) {
      if (item instanceof Wall
          || item instanceof Room
          || item instanceof HomePieceOfFurniture) {
        otherVisibleHomeItems = true;
        break;
      }
    }
    createComponents(editedRooms, otherVisibleHomeItems, resource);

    final JOptionPane optionPane = new JOptionPane(this.terrainPanel, JOptionPane.PLAIN_MESSAGE, JOptionPane.OK_CANCEL_OPTION);
    final JDialog dialog = optionPane.createDialog((JComponent)homeView, resource.getString("terrainGeneratorTitle"));
    dialog.applyComponentOrientation(((JComponent)homeView).getComponentOrientation());
    dialog.setLocationByPlatform(true);
    dialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
    dialog.setResizable(true);
    dialog.setVisible(true);

    if (optionPane.getValue() != null
        && optionPane.getValue().equals(JOptionPane.OK_OPTION)) {
      // Create a new 3D model in a threaded task
      Callable<Void> exportToObjTask = new Callable<Void>() {
          public Void call() throws RecorderException {
            final String terrainName = resource.getString("untitledTerrainName");
            final Content modelContent = terrainPreview.exportTerrain(terrainName);
            float minX = Float.POSITIVE_INFINITY;
            float maxX = Float.NEGATIVE_INFINITY;
            float minY = Float.POSITIVE_INFINITY;
            float maxY = Float.NEGATIVE_INFINITY;
            for (Room room : selectedRooms) {
              for (float [] point : room.getPoints()) {
                minX = Math.min(minX, point [0]);
                maxX = Math.max(maxX, point [0]);
                minY = Math.min(minY, point [1]);
                maxY = Math.max(maxY, point [1]);
              }
            }
            final float x = (minX + maxX) / 2;
            final float y = (minY + maxY) / 2;
            EventQueue.invokeLater(new Runnable() {
                public void run() {
                  if (modelContent != null) {
                    addNewPieceOfFurniture(modelContent, terrainName, x, y, 0, selectedRooms, editedRooms, resource);
                  } else {
                    JOptionPane.showMessageDialog((JComponent)homeView,
                        resource.getString("emptyTerrainError.message"),
                        resource.getString("terrainGeneratorError.title"), JOptionPane.ERROR_MESSAGE);
                  }
                }
              });
            return null;
          }
        };
      ThreadedTaskController.ExceptionHandler exceptionHandler =
          new ThreadedTaskController.ExceptionHandler() {
            public void handleException(Exception ex) {
              if (!(ex instanceof InterruptedRecorderException)) {
                showError((JComponent)homeView, resource, ex.getMessage());
                ex.printStackTrace();
              }
            }
          };
      new ThreadedTaskController(exportToObjTask,
          resource.getString("terrainGeneratorMessage"), exceptionHandler,
          getUserPreferences(), new SwingViewFactory()).executeTask(homeView);
    }
  }

  /**
   * Creates and layout the components shown in the dialog box of this plug-in.
   */
  private void createComponents(final List<Room> editedRooms,
                                boolean otherVisibleHomeItems,
                                final ResourceBundle resource) {
    UserPreferences preferences = getUserPreferences();

    JLabel infoLabel = new JLabel(" ");

    this.terrainPreview = new TerrainPreviewComponent(getHome(), getUserPreferences(), infoLabel, resource);
    this.terrainPreview.setFocusable(false);
    this.terrainPreview.setBackground(Color.LIGHT_GRAY);
    this.terrainPreview.setTerrainRooms(editedRooms);

    JButton resetButton = new JButton(new AbstractAction(resource.getString("resetButton.text")) {
        public void actionPerformed(ActionEvent ev) {
          resetTerrainElevationProperties(editedRooms);
          Group terrainsGroup = terrainPreview.getTerrainsGroup();
          if (terrainsGroup != null) {
            for (int i = 0, n = terrainsGroup.numChildren(); i < n; i++) {
              ((Terrain3D)terrainsGroup.getChild(i)).update();
            }
          }
        }
      });

    JLabel influenceRadiusLabel = new JLabel(String.format(resource.getString("influenceRadiusLabel.text"), getUserPreferences().getLengthUnit().getName()));
    final SpinnerNumberModel influenceRadiusSpinnerModel = new SpinnerNumberModel(new Float(this.terrainPreview.getInfluenceRadius()), new Float(50f), new Float(100000f), new Float(100f));
    JSpinner influenceRadiusSpinner = new AutoCommitSpinner(influenceRadiusSpinnerModel);
    influenceRadiusSpinnerModel.addChangeListener(new ChangeListener() {
        public void stateChanged(ChangeEvent ev) {
          terrainPreview.setInfluenceRadius(((Number)influenceRadiusSpinnerModel.getValue()).floatValue());
        }
      });

    JCheckBox displayHomeCheckBox = new JCheckBox(new AbstractAction(resource.getString("displayHomeCheckBox.text")) {
        public void actionPerformed(ActionEvent ev) {
          terrainPreview.setHomeVisible(((JCheckBox)ev.getSource()).isSelected());
        }
      });
    displayHomeCheckBox.setSelected(true);

    // Layout coordinates spinners and transparency slider in a panel with labels
    this.terrainPanel = new JPanel(new GridBagLayout());
    this.terrainPanel.add(new JLabel(String.format(resource.getString("pointsPanel.comment"), preferences.getLengthUnit().getName())),
        new GridBagConstraints(0, 0, 2, 1, 0, 0, GridBagConstraints.LINE_START,
            GridBagConstraints.NONE, new Insets(0, 0, 5, 5), 0, 0));
    this.terrainPanel.add(resetButton,
        new GridBagConstraints(1, 0, 2, 1, 0, 0, GridBagConstraints.LINE_END,
            GridBagConstraints.NONE, new Insets(0, 0, 5, 0), 0, 0));
    this.terrainPanel.add(this.terrainPreview,
        new GridBagConstraints(0, 1, 3, 1, 1, 1, GridBagConstraints.LINE_START,
            GridBagConstraints.BOTH, new Insets(0, 0, 0, 0), 0, 0));
    this.terrainPanel.add(infoLabel,
        new GridBagConstraints(0, 2, 1, 2, 1, 0, GridBagConstraints.LINE_START,
            GridBagConstraints.NONE, new Insets(5, 0, 0, 10), 0, 0));
    this.terrainPanel.add(influenceRadiusLabel,
        new GridBagConstraints(1, 2, 1, 1, 0, 0, GridBagConstraints.LINE_END,
            GridBagConstraints.NONE, new Insets(5, 0, 0, 5), 0, 0));
    this.terrainPanel.add(influenceRadiusSpinner,
        new GridBagConstraints(2, 2, 1, 1, 0, 0, GridBagConstraints.LINE_END,
            GridBagConstraints.NONE, new Insets(5, 0, 0, 0), 0, 0));
    this.terrainPanel.add(otherVisibleHomeItems ? displayHomeCheckBox : new JLabel(" "),
        new GridBagConstraints(1, 3, 2, 1, 0, 0, GridBagConstraints.LINE_END,
            GridBagConstraints.NONE, new Insets(5, 0, 0, 0), 0, 0));

    if (!OperatingSystem.isMacOSX()) {
      resetButton.setMnemonic(KeyStroke.getKeyStroke(resource.getString("resetButton.mnemonic")).getKeyCode());
      displayHomeCheckBox.setMnemonic(KeyStroke.getKeyStroke(resource.getString("displayHomeCheckBox.mnemonic")).getKeyCode());
      influenceRadiusLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(resource.getString("influenceRadiusLabel.mnemonic")).getKeyCode());
      influenceRadiusLabel.setLabelFor(influenceRadiusSpinner);
    }
  }

  /**
   * Resets all elevations of the given <code>rooms</code>.
   */
  private static void resetTerrainElevationProperties(List<Room> rooms) {
    for (Room room : rooms) {
      for (String propertyName : new ArrayList<String>(room.getPropertyNames())) {
        if (propertyName.startsWith(Terrain3D.TERRAIN_ELEVATION_KEY_PREFIX)) {
          room.setProperty(propertyName, null);
        }
      }
      room.setProperty(Terrain3D.TERRAIN_GRID_SIZE_KEY, null);
    }
  }

  /**
   * Adds to home a new piece of furniture based on the given model.
   */
  private void addNewPieceOfFurniture(final Content modelContent,
                                      final String terrainName,
                                      final float x,
                                      final float y,
                                      final float elevation,
                                      final List<Room> selectedRooms,
                                      final List<Room> editedRooms,
                                      final ResourceBundle resource) {
    ModelManager.getInstance().loadModel(modelContent, new ModelManager.ModelObserver() {
        public void modelUpdated(BranchGroup modelRoot) {
          Vector3f size = ModelManager.getInstance().getSize(modelRoot);
          Content iconContent = new ResourceURLContent(RoomsTerrainGeneratorPlugin.class, "terrain.png");
          HomePieceOfFurniture piece = new HomePieceOfFurniture(new CatalogPieceOfFurniture(
                null, terrainName, null, null, null, null, null, iconContent, null, modelContent,
                size.x, size.z, size.y, 0, 0,
                false, null, null, false, null, System.getProperty("user.name"),
                true, true, true, false, // Not horizontally rotatable
                null, null, null, null));
          piece.setX(x);
          piece.setY(y);
          piece.setElevation(elevation);
          final List<Map<String, String>> oldElevations = new ArrayList<Map<String, String>>(selectedRooms.size());
          final String [] oldGridSizes = new String [selectedRooms.size()];
          for (int i = 0; i < selectedRooms.size(); i++) {
            Room room = selectedRooms.get(i);
            oldElevations.add(new HashMap<String, String>());
            for (String propertyName : room.getPropertyNames()) {
              if (propertyName.startsWith(Terrain3D.TERRAIN_ELEVATION_KEY_PREFIX)) {
                oldElevations.get(i).put(propertyName, room.getProperty(propertyName));
              }
            }
            oldGridSizes [i] = room.getProperty(Terrain3D.TERRAIN_GRID_SIZE_KEY);
          }
          resetTerrainElevationProperties(selectedRooms);
          final List<Map<String, String>> newElevations = new ArrayList<Map<String, String>>(selectedRooms.size());
          final String [] newGridSizes = new String [selectedRooms.size()];
          for (int i = 0; i < editedRooms.size(); i++) {
            Room room = editedRooms.get(i);
            newElevations.add(new HashMap<String, String>());
            for (String propertyName : room.getPropertyNames()) {
              if (propertyName.startsWith(Terrain3D.TERRAIN_ELEVATION_KEY_PREFIX)) {
                String value = room.getProperty(propertyName);
                newElevations.get(i).put(propertyName, value);
                selectedRooms.get(i).setProperty(propertyName, value);
              }
            }
            newGridSizes [i] = room.getProperty(Terrain3D.TERRAIN_GRID_SIZE_KEY);
            selectedRooms.get(i).setProperty(Terrain3D.TERRAIN_GRID_SIZE_KEY, newGridSizes [i]);
          }
          UndoableEditSupport undoSupport = getUndoableEditSupport();
          undoSupport.beginUpdate();
          getHomeController().getFurnitureController().addFurniture(Arrays.asList(new HomePieceOfFurniture [] {piece}));
          undoSupport.postEdit(new AbstractUndoableEdit() {
              @Override
              public void undo() throws CannotUndoException {
                super.undo();
                resetTerrainElevationProperties(selectedRooms);
                for (int i = 0; i < selectedRooms.size(); i++) {
                  Room room = selectedRooms.get(i);
                  for (Map.Entry<String, String> entry : oldElevations.get(i).entrySet()) {
                    room.setProperty(entry.getKey(), entry.getValue());
                  }
                  room.setProperty(Terrain3D.TERRAIN_GRID_SIZE_KEY, oldGridSizes [i]);
                }
              }

              @Override
              public void redo() throws CannotRedoException {
                super.redo();
                resetTerrainElevationProperties(selectedRooms);
                for (int i = 0; i < selectedRooms.size(); i++) {
                  Room room = selectedRooms.get(i);
                  for (Map.Entry<String, String> entry : newElevations.get(i).entrySet()) {
                    room.setProperty(entry.getKey(), entry.getValue());
                  }
                  room.setProperty(Terrain3D.TERRAIN_GRID_SIZE_KEY, newGridSizes [i]);
                }
              }

              @Override
              public String getPresentationName() {
                return resource.getString("undoGenerateTerrain");
              }
            });

          // End compound edit
          undoSupport.endUpdate();
        }

        public void modelError(Exception ex) {
          // Shouldn't happen since we import a model we just exported
        }
      });
  }

  /**
   * Shows a message error.
   */
  private void showError(Component parent,
                         ResourceBundle resource,
                         String messageDetail) {
    String messageFormat = resource.getString("terrainGeneratorError.message");
    JOptionPane.showMessageDialog(parent, String.format(messageFormat, messageDetail),
        resource.getString("terrainGeneratorError.title"), JOptionPane.ERROR_MESSAGE);
  }

  // Inspired from ModelPreviewComponent
  private static class TerrainPreviewComponent extends JComponent {
    private static final int MODEL_PREFERRED_SIZE = Math.round(600 * SwingTools.getResolutionScale());

    private UserPreferences  preferences;
    private JLabel           infoLabel;
    private ResourceBundle   resource;
    private boolean          magnetismEnabled;
    private boolean          alignmentActivated;
    private float            influenceRadius = Terrain3D.DEFAULT_GRID_SIZE * 10;
    private SimpleUniverse   universe;
    private JPanel           component3DPanel;
    private Component        component3D;
    private BranchGroup      sceneTree;
    private float            viewYaw   = 0;
    private float            viewPitch = -(float)Math.PI / 4;
    private float            viewScale = 1;
    private boolean          parallelProjection;
    private List<Room>       previewedRooms;
    private Point3d          pickedVertex;

    /**
     * Returns an 3D model preview component that lets the user change its yaw, pitch and scale
     * according to parameters.
     */
    public TerrainPreviewComponent(Home home, final UserPreferences preferences, JLabel infoLabel, ResourceBundle resource) {
      setBorder(BorderFactory.createEtchedBorder(EtchedBorder.LOWERED));

      this.preferences = preferences;
      this.infoLabel = infoLabel;
      this.resource = resource;
      this.sceneTree = createSceneTree(home);

      this.component3DPanel = new JPanel();
      setLayout(new BorderLayout());
      add(this.component3DPanel);

      GraphicsEnvironment graphicsEnvironment = GraphicsEnvironment.getLocalGraphicsEnvironment();
      if (graphicsEnvironment.getScreenDevices().length == 1) {
        // If only one screen device is available, create 3D component immediately,
        // otherwise create it once the screen device of the parent is known
        createComponent3D(graphicsEnvironment.getDefaultScreenDevice().getDefaultConfiguration());
      }

      // Add an ancestor listener to create 3D component and its universe once this component is made visible
      // and clean up universe once its parent frame is disposed
      addAncestorListener();

      this.magnetismEnabled = preferences.isMagnetismEnabled();
      ActionMap actionMap = getActionMap();
      actionMap.put("TOGGLE_MAGNETISM_ON", new AbstractAction() {
          public void actionPerformed(ActionEvent ev) {
            magnetismEnabled = !preferences.isMagnetismEnabled();
            updateInfo(null);
          }
        });
      actionMap.put("TOGGLE_MAGNETISM_OFF", new AbstractAction() {
          public void actionPerformed(ActionEvent ev) {
            magnetismEnabled = preferences.isMagnetismEnabled();
            updateInfo(null);
          }
        });
      actionMap.put("ACTIVATE_ALIGNMENT", new AbstractAction() {
          public void actionPerformed(ActionEvent ev) {
            alignmentActivated = true;
            updateInfo(null);
          }
        });
      actionMap.put("DEACTIVATE_ALIGNMENT", new AbstractAction() {
          public void actionPerformed(ActionEvent ev) {
            alignmentActivated = false;
            updateInfo(null);
          }
        });

      InputMap inputMap = getInputMap(WHEN_IN_FOCUSED_WINDOW);
      if (OperatingSystem.isMacOSX()) {
        inputMap.put(KeyStroke.getKeyStroke("meta pressed META"), "TOGGLE_MAGNETISM_ON");
        inputMap.put(KeyStroke.getKeyStroke("released META"), "TOGGLE_MAGNETISM_OFF");
        inputMap.put(KeyStroke.getKeyStroke("shift meta pressed META"), "TOGGLE_MAGNETISM_ON");
        inputMap.put(KeyStroke.getKeyStroke("shift released META"), "TOGGLE_MAGNETISM_OFF");
        inputMap.put(KeyStroke.getKeyStroke("shift pressed SHIFT"), "ACTIVATE_ALIGNMENT");
        inputMap.put(KeyStroke.getKeyStroke("released SHIFT"), "DEACTIVATE_ALIGNMENT");
        inputMap.put(KeyStroke.getKeyStroke("shift meta pressed SHIFT"), "ACTIVATE_ALIGNMENT");
        inputMap.put(KeyStroke.getKeyStroke("meta released SHIFT"), "DEACTIVATE_ALIGNMENT");
      } else {
        inputMap.put(KeyStroke.getKeyStroke("alt pressed ALT"), "TOGGLE_MAGNETISM_ON");
        inputMap.put(KeyStroke.getKeyStroke("released ALT"), "TOGGLE_MAGNETISM_OFF");
        inputMap.put(KeyStroke.getKeyStroke("shift meta pressed ALT"), "TOGGLE_MAGNETISM_ON");
        inputMap.put(KeyStroke.getKeyStroke("shift released ALT"), "TOGGLE_MAGNETISM_OFF");
        inputMap.put(KeyStroke.getKeyStroke("shift pressed SHIFT"), "ACTIVATE_ALIGNMENT");
        inputMap.put(KeyStroke.getKeyStroke("released SHIFT"), "DEACTIVATE_ALIGNMENT");
        inputMap.put(KeyStroke.getKeyStroke("shift alt pressed SHIFT"), "ACTIVATE_ALIGNMENT");
        inputMap.put(KeyStroke.getKeyStroke("alt released SHIFT"), "DEACTIVATE_ALIGNMENT");
      }
    }

    /**
     * Returns component preferred size.
     */
    @Override
    public Dimension getPreferredSize() {
      if (isPreferredSizeSet()) {
        return super.getPreferredSize();
      } else {
        return new Dimension(MODEL_PREFERRED_SIZE, MODEL_PREFERRED_SIZE * 2 / 3);
      }
    }

    @Override
    public void addMouseMotionListener(final MouseMotionListener l) {
      super.addMouseMotionListener(l);
      if (this.component3D != null) {
        this.component3D.addMouseMotionListener(new MouseMotionListener() {
            public void mouseMoved(MouseEvent ev) {
              l.mouseMoved(SwingUtilities.convertMouseEvent(component3D, ev, TerrainPreviewComponent.this));
            }

            public void mouseDragged(MouseEvent ev) {
              l.mouseDragged(SwingUtilities.convertMouseEvent(component3D, ev, TerrainPreviewComponent.this));
            }
          });
      }
    }

    @Override
    public void addMouseListener(final MouseListener l) {
      super.addMouseListener(l);
      if (this.component3D != null) {
        this.component3D.addMouseListener(new MouseListener() {
            public void mouseReleased(MouseEvent ev) {
              l.mouseReleased(SwingUtilities.convertMouseEvent(component3D, ev, TerrainPreviewComponent.this));
            }

            public void mousePressed(MouseEvent ev) {
              l.mousePressed(SwingUtilities.convertMouseEvent(component3D, ev, TerrainPreviewComponent.this));
            }

            public void mouseExited(MouseEvent ev) {
              l.mouseExited(SwingUtilities.convertMouseEvent(component3D, ev, TerrainPreviewComponent.this));
            }

            public void mouseEntered(MouseEvent ev) {
              l.mouseEntered(SwingUtilities.convertMouseEvent(component3D, ev, TerrainPreviewComponent.this));
            }

            public void mouseClicked(MouseEvent ev) {
              l.mouseClicked(SwingUtilities.convertMouseEvent(component3D, ev, TerrainPreviewComponent.this));
            }
          });
      }
    }

    /**
     * Adds an ancestor listener to this component to manage the creation of the 3D component and its universe
     * and clean up the universe.
     */
    private void addAncestorListener() {
      addAncestorListener(new AncestorListener() {
          public void ancestorAdded(AncestorEvent ev) {
            if (component3D == null) {
              createComponent3D(ev.getAncestor().getGraphicsConfiguration());
            }
            if (universe == null) {
              createUniverse();
            }
          }

          public void ancestorRemoved(AncestorEvent ev) {
            if (universe != null) {
              // Dispose universe later to avoid conflicts if canvas is currently being redrawn
              EventQueue.invokeLater(new Runnable() {
                  public void run() {
                    disposeUniverse();
                    component3DPanel.removeAll();
                    component3D = null;
                  }
                });
            }
          }

          public void ancestorMoved(AncestorEvent ev) {
          }
        });
    }

    /**
     * Creates the 3D component associated with the given <code>configuration</code> device.
     */
    private void createComponent3D(GraphicsConfiguration graphicsConfiguration) {
      if (Boolean.getBoolean("com.eteks.sweethome3d.j3d.useOffScreen3DView")) {
        GraphicsConfigTemplate3D gc = new GraphicsConfigTemplate3D();
        gc.setSceneAntialiasing(GraphicsConfigTemplate3D.PREFERRED);
        try {
          // Instantiate JCanvas3DWithNotifiedPaint inner class by reflection
          // to be able to run under Java 3D 1.3
          this.component3D = (Component)Class.forName(TerrainPreviewComponent.class.getName() + "$JCanvas3DWithNotifiedPaint").
              getConstructor(TerrainPreviewComponent.class, GraphicsConfigTemplate3D.class).newInstance(this, gc);
        } catch (ClassNotFoundException ex) {
          throw new UnsupportedOperationException("Java 3D 1.5 required to display an offscreen 3D view");
        } catch (Exception ex) {
          UnsupportedOperationException ex2 = new UnsupportedOperationException();
          ex2.initCause(ex);
          throw ex2;
        }
      } else {
        this.component3D = Component3DManager.getInstance().getOnscreenCanvas3D(graphicsConfiguration, null);
      }

      // Layout 3D component
      this.component3DPanel.setLayout(new GridLayout());
      this.component3DPanel.add(this.component3D);
      this.component3D.setFocusable(false);
      addMouseListeners(this.component3D);
    }

    /**
     * A <code>JCanvas</code> canvas that sends a notification when it's drawn.
     */
    private static class JCanvas3DWithNotifiedPaint extends JCanvas3D {
      private final TerrainPreviewComponent homeComponent3D;

      public JCanvas3DWithNotifiedPaint(TerrainPreviewComponent component,
                                        GraphicsConfigTemplate3D template) {
        super(template);
        this.homeComponent3D = component;
      }

      public void paintComponent(Graphics g) {
        super.paintComponent(g);
      }
    }

    /**
     * Adds an AWT mouse listener to component that will update view platform transform.
     */
    private void addMouseListeners(final Component component3D) {
      final float ANGLE_FACTOR = 0.02f;
      final float ZOOM_FACTOR = 0.02f;

      MouseInputAdapter mouseListener = new MouseInputAdapter() {
          private int            xLastMouseMove;
          private int            yLastMouseMove;
          private boolean        escaped;
          private boolean        aligned;
          private boolean        magnetized;
          private Transform3D    transformationToScene;
          private double         elevationMousePressed;
          private List<Map<String, String>> terrainElevationsMousePressed = null;

          private Point getMouseLocation(MouseEvent ev) {
            if (OperatingSystem.isJavaVersionGreaterOrEqual("1.9")) {
              try {
                // Dirty hack that scales mouse coordinates with xcale and yscale private fields of Canvas3D
                Field xscaleField = Canvas3D.class.getDeclaredField("xscale");
                xscaleField.setAccessible(true);
                double xscale = (Double)(xscaleField.get(ev.getSource()));
                Field yscaleField = Canvas3D.class.getDeclaredField("yscale");
                yscaleField.setAccessible(true);
                double yscale = (Double)(yscaleField.get(ev.getSource()));
                return new Point((int)(ev.getX() * xscale), (int)(ev.getY() * yscale));
              } catch (Exception ex) {
              }
            }
            return ev.getPoint();
          }

          @Override
          public void mousePressed(MouseEvent ev) {
            Point mouseLocation = getMouseLocation(ev);
            this.xLastMouseMove = mouseLocation.x;
            this.yLastMouseMove = mouseLocation.y;
            this.escaped = false;
            this.aligned = alignmentActivated;
            this.magnetized = magnetismEnabled;
            if (getTerrainsGroup() != null) {
              Canvas3D canvas;
              if (component3D instanceof Canvas3D) {
                canvas = (Canvas3D)component3D;
              } else {
                canvas = ((JCanvas3D)component3D).getOffscreenCanvas3D();
              }
              PickCanvas pickCanvas = new PickCanvas(canvas, sceneTree);
              pickCanvas.setMode(PickCanvas.GEOMETRY);
              pickCanvas.setShapeLocation(mouseLocation.x, mouseLocation.y);
              PickResult result = pickCanvas.pickClosest();
              PickIntersection intersection;
              if (result != null
                  && (intersection = result.getClosestIntersection(pickCanvas.getStartPosition())) != null) {
                pickedVertex = intersection.getClosestVertexCoordinates();
                Node referenceNode = result.getNode(PickResult.SHAPE3D);
                Transform3D transform = getTransformBetweenNodes(referenceNode.getParent(), sceneTree);
                this.transformationToScene = new Transform3D();
                canvas.getVworldToImagePlate(this.transformationToScene);
                this.transformationToScene.mul(transform);
                this.transformationToScene.invert();

                Point3d point = new Point3d();
                canvas.getPixelLocationInImagePlate(mouseLocation.x, mouseLocation.y, point);
                this.transformationToScene.transform(point);
                this.elevationMousePressed = point.y;
                if (this.terrainElevationsMousePressed == null) {
                  this.terrainElevationsMousePressed = new ArrayList<Map<String,String>>();
                  for (int i = 0; i < previewedRooms.size(); i++) {
                    this.terrainElevationsMousePressed.add(new HashMap<String, String>());
                  }
                } else {
                  for (Map<String,String> map : this.terrainElevationsMousePressed) {
                    map.clear();
                  }
                }
                for (int i = 0; i < previewedRooms.size(); i++) {
                  Room room = previewedRooms.get(i);
                  for (String propertyName : room.getPropertyNames()) {
                    if (propertyName.startsWith(Terrain3D.TERRAIN_ELEVATION_KEY_PREFIX)) {
                      this.terrainElevationsMousePressed.get(i).put(propertyName, room.getProperty(propertyName));
                    }
                  }
                }

                setFeedbackLocation((float)pickedVertex.x, (float)pickedVertex.y, (float)pickedVertex.z, 0, false);
                updateInfo(pickedVertex.y);
                getActionMap().put("ESCAPE_TERRAIN_MODIFICATION", new AbstractAction() {
                    public void actionPerformed(ActionEvent ev) {
                      escape();
                    }
                  });
                InputMap inputMap = getInputMap(WHEN_IN_FOCUSED_WINDOW);
                inputMap.put(KeyStroke.getKeyStroke("ESCAPE"), "ESCAPE_TERRAIN_MODIFICATION");
                inputMap.put(KeyStroke.getKeyStroke("shift ESCAPE"), "ESCAPE_TERRAIN_MODIFICATION");
                inputMap.put(KeyStroke.getKeyStroke("alt ESCAPE"), "ESCAPE_TERRAIN_MODIFICATION");
              } else {
                transformationToScene = null;
                infoLabel.setText(" ");
              }
            }
          }

          private Transform3D getTransformBetweenNodes(Node node, Node parent) {
            Transform3D transform = new Transform3D();
            if (node instanceof TransformGroup) {
              ((TransformGroup)node).getTransform(transform);
            }
            if (node != parent) {
              Node nodeParent = node.getParent();
              if (nodeParent instanceof Group) {
                transform.mul(getTransformBetweenNodes(nodeParent, parent), transform);
              } else {
                throw new IllegalStateException("Can't retrieve node transform");
              }
            }
            return transform;
          }

          @Override
          public void mouseDragged(MouseEvent ev) {
            Point mouseLocation = getMouseLocation(ev);
            Group terrainsGroup = getTerrainsGroup();
            if (terrainsGroup != null
                && !this.escaped) {
              if (this.transformationToScene != null) {
                if (this.yLastMouseMove != mouseLocation.y) {
                  if (aligned != alignmentActivated
                      || magnetized != magnetismEnabled) {
                    resetTerrainElevationPropertiesAtMousePress();
                    aligned = alignmentActivated;
                    magnetized = magnetismEnabled;
                  }

                  Point3d point = new Point3d();
                  Canvas3D canvas;
                  if (component3D instanceof Canvas3D) {
                    canvas = (Canvas3D)component3D;
                  } else {
                    canvas = ((JCanvas3D)component3D).getOffscreenCanvas3D();
                  }
                  canvas.getPixelLocationInImagePlate(mouseLocation.x, mouseLocation.y, point);
                  this.transformationToScene.transform(point);
                  float elevationDelta = (float)(point.y - this.elevationMousePressed);

                  Set<String> updatedProperties = new HashSet<String>();
                  for (int i = 0, n = terrainsGroup.numChildren(); i < n; i++) {
                    Terrain3D terrainNode = (Terrain3D)terrainsGroup.getChild(i);
                    Room previewedRoom = previewedRooms.get(i);
                    float gridSize = Terrain3D.DEFAULT_GRID_SIZE;
                    String gridSizeValue = previewedRoom.getProperty(Terrain3D.TERRAIN_GRID_SIZE_KEY);
                      if (gridSizeValue != null) {
                      try {
                        gridSize = Float.parseFloat(gridSizeValue);
                      } catch (NumberFormatException ex) {
                      }
                    }
                    Map<String, String> terrainElevations = this.terrainElevationsMousePressed.get(i);
                    updatedProperties.clear();
                    for (Enumeration<Geometry> enumGeometry = terrainNode.getTerrainShape().getAllGeometries(); enumGeometry.hasMoreElements(); ) {
                      Geometry geometry = enumGeometry.nextElement();
                      if (geometry instanceof GeometryArray) {
                        GeometryArray geometryArray = (GeometryArray)geometry;
                        int vertexCount = geometryArray.getVertexCount();
                        Point3f vertex = new Point3f();
                        if ((geometryArray.getVertexFormat() & GeometryArray.BY_REFERENCE) != 0) {
                          if ((geometryArray.getVertexFormat() & GeometryArray.INTERLEAVED) != 0) {
                            float [] vertexData = geometryArray.getInterleavedVertices();
                            int vertexSize = vertexData.length / vertexCount;
                            for (int index = 0, j = vertexSize - 3; index < vertexCount; j += vertexSize, index++) {
                              updateVertexElevation(vertexData [j], vertexData [j + 1], vertexData [j + 2],
                                  elevationDelta, Terrain3D.DEFAULT_GRID_SIZE, previewedRoom, terrainElevations, updatedProperties);
                            }
                          } else {
                            float [] vertexCoordinates = geometryArray.getCoordRefFloat();
                            for (int index = 0, j = 0; index < vertexCount; j += 3, index++) {
                              updateVertexElevation(vertexCoordinates [j], vertexCoordinates [j + 1], vertexCoordinates [j + 2],
                                   elevationDelta, Terrain3D.DEFAULT_GRID_SIZE, previewedRoom, terrainElevations, updatedProperties);
                            }
                          }
                        } else {
                          for (int index = 0, j = 0; index < vertexCount; j++, index++) {
                            geometryArray.getCoordinate(j, vertex);
                            updateVertexElevation(vertex.x, vertex.y, vertex.z,
                                elevationDelta, Terrain3D.DEFAULT_GRID_SIZE, previewedRoom, terrainElevations, updatedProperties);
                          }
                        }
                      }
                    }
                    if (updatedProperties.size() > 0) {
                      previewedRoom.setProperty(Terrain3D.TERRAIN_GRID_SIZE_KEY, Terrain3D.TERRAIN_GRID_SIZE_FORMAT.format(gridSize));
                      terrainNode.update();
                    }
                  }
                }

                updateInfo(null);
              } else {
                if (ev.isAltDown()) {
                  // Mouse move along Y axis with Alt down changes scale
                  setViewScale(Math.max(0.5f, Math.min(1.7f, getViewScale() * (float)Math.exp((mouseLocation.y - this.yLastMouseMove) * ZOOM_FACTOR))));
                } else if (!ev.isAltDown()) {
                  // Mouse move along X axis changes yaw
                  setViewYaw(getViewYaw() - ANGLE_FACTOR * (mouseLocation.x - this.xLastMouseMove));
                  // Mouse move along Y axis changes pitch
                  float viewPitch = getViewPitch() - ANGLE_FACTOR * (mouseLocation.y - this.yLastMouseMove);
                  setViewPitch(Math.max(-(float)Math.PI / 2, Math.min(0, viewPitch)));
                }
              }
            }
            this.xLastMouseMove = mouseLocation.x;
            this.yLastMouseMove = mouseLocation.y;
          }

          @Override
          public void mouseReleased(MouseEvent ev) {
            InputMap inputMap = getInputMap(WHEN_IN_FOCUSED_WINDOW);
            inputMap.remove(KeyStroke.getKeyStroke("ESCAPE"));
            inputMap.remove(KeyStroke.getKeyStroke("alt ESCAPE"));
            inputMap.remove(KeyStroke.getKeyStroke("shift ESCAPE"));
            setFeedbackLocation(Float.NaN, 0, 0, 0, false);
          }

          private void updateVertexElevation(float x, float y, float z, float elevationDelta, float gridSize,
                                             Room updatedRoom, Map<String, String> terrainElevations, Set<String> updatedProperties) {
            float distanceToPickedVertex = (float)(alignmentActivated && magnetismEnabled
                ? Line2D.ptLineDist(pickedVertex.x, pickedVertex.z, pickedVertex.x + Math.cos(getViewYaw()), pickedVertex.z - Math.sin(getViewYaw()), x, z)
                : Point2D.distance(pickedVertex.x, pickedVertex.z, x, z));
            float deviationSquare = (influenceRadius / 10) * (influenceRadius / 10) * 6;
            float vertexElevationDelta = normalDistribution(distanceToPickedVertex, deviationSquare) * elevationDelta;
            if ((!magnetismEnabled && distanceToPickedVertex == 0)
                || (magnetismEnabled && Math.abs(vertexElevationDelta) >= 1)) {
              String vertexElevationPropertyKey = String.format(Locale.US, Terrain3D.TERRAIN_ELEVATION_KEY_FORMAT, x, z);
              if (!updatedProperties.contains(vertexElevationPropertyKey)) {
                float elevation = vertexElevationDelta;
                String oldElevation = terrainElevations.get(vertexElevationPropertyKey);
                if (oldElevation != null) {
                  try {
                    elevation = Float.parseFloat(oldElevation) + vertexElevationDelta;
                  } catch (NumberFormatException ex) {
                  }
                }
                updatedRoom.setProperty(vertexElevationPropertyKey,
                    elevation <= 0 ? null : Terrain3D.TERRAIN_ELEVATION_FORMAT.format(Math.max(elevation, 0)));
                // Update elevation only once
                updatedProperties.add(vertexElevationPropertyKey);

                if (x == pickedVertex.x && z == pickedVertex.z) {
                  setFeedbackLocation((float)pickedVertex.x, Math.max(0, elevation), (float)pickedVertex.z,
                      magnetismEnabled ? (float)Math.sqrt(-Math.log(1 / Math.abs(elevationDelta)) * 2 * deviationSquare) : 0,
                      alignmentActivated && magnetismEnabled);
                }
              }
            }
          }

          private float normalDistribution(float x, float deviationSquare) {
            return (float)Math.exp(-x * x / 2 / deviationSquare);
          }

          private void escape() {
            resetTerrainElevationPropertiesAtMousePress();
            Group terrainsGroup = getTerrainsGroup();
            for (int i = 0, n = terrainsGroup.numChildren(); i < n; i++) {
              ((Terrain3D)terrainsGroup.getChild(i)).update();
            }
            this.escaped = true;
            infoLabel.setText(" ");
            mouseReleased(null);
          }

          private void resetTerrainElevationPropertiesAtMousePress() {
            resetTerrainElevationProperties(previewedRooms);
            for (int i = 0; i < previewedRooms.size(); i++) {
              Room room = previewedRooms.get(i);
              for (Entry<String, String> property : terrainElevationsMousePressed.get(i).entrySet()) {
                room.setProperty(property.getKey(), property.getValue());
              }
            }
          }
        };

      component3D.addMouseListener(mouseListener);
      component3D.addMouseMotionListener(mouseListener);

      component3D.addMouseWheelListener(new MouseWheelListener() {
          public void mouseWheelMoved(MouseWheelEvent ev) {
            // Mouse move along Y axis with Alt down changes scale
            setViewScale(Math.max(0.5f, Math.min(1.7f, getViewScale() * (float)Math.exp(ev.getWheelRotation() * ZOOM_FACTOR))));
          }
        });

      // Redirect mouse events to the 3D component
      for (final MouseMotionListener l : getListeners(MouseMotionListener.class)) {
        component3D.addMouseMotionListener(new MouseMotionListener() {
            public void mouseMoved(MouseEvent ev) {
              l.mouseMoved(SwingUtilities.convertMouseEvent(component3D, ev, TerrainPreviewComponent.this));
            }

            public void mouseDragged(MouseEvent ev) {
              l.mouseDragged(SwingUtilities.convertMouseEvent(component3D, ev, TerrainPreviewComponent.this));
            }
          });
      }
      for (final MouseListener l : getListeners(MouseListener.class)) {
        component3D.addMouseListener(new MouseListener() {
            public void mouseReleased(MouseEvent ev) {
              l.mouseReleased(SwingUtilities.convertMouseEvent(component3D, ev, TerrainPreviewComponent.this));
            }

            public void mousePressed(MouseEvent ev) {
              l.mousePressed(SwingUtilities.convertMouseEvent(component3D, ev, TerrainPreviewComponent.this));
            }

            public void mouseExited(MouseEvent ev) {
              l.mouseExited(SwingUtilities.convertMouseEvent(component3D, ev, TerrainPreviewComponent.this));
            }

            public void mouseEntered(MouseEvent ev) {
              l.mouseEntered(SwingUtilities.convertMouseEvent(component3D, ev, TerrainPreviewComponent.this));
            }

            public void mouseClicked(MouseEvent ev) {
              l.mouseClicked(SwingUtilities.convertMouseEvent(component3D, ev, TerrainPreviewComponent.this));
            }
          });
      }
    }

    /**
     * Creates universe bound to the 3D component.
     */
    private void createUniverse() {
      Canvas3D canvas3D;
      if (this.component3D instanceof Canvas3D) {
        canvas3D = (Canvas3D)this.component3D;
      } else {
        try {
          // Call JCanvas3D#getOffscreenCanvas3D by reflection to be able to run under Java 3D 1.3
          canvas3D = (Canvas3D)Class.forName("com.sun.j3d.exp.swing.JCanvas3D").getMethod("getOffscreenCanvas3D").invoke(this.component3D);
        } catch (Exception ex) {
          UnsupportedOperationException ex2 = new UnsupportedOperationException();
          ex2.initCause(ex);
          throw ex2;
        }
      }
      // Create a universe bound to component 3D
      ViewingPlatform viewingPlatform = new ViewingPlatform();
      Viewer viewer = new Viewer(canvas3D);
      this.universe = new SimpleUniverse(viewingPlatform, viewer);
      // Link scene to universe
      this.universe.addBranchGraph(this.sceneTree);
      this.universe.getViewer().getView().setProjectionPolicy(this.parallelProjection
          ? View.PARALLEL_PROJECTION
          : View.PERSPECTIVE_PROJECTION);
      // Set viewer location
      updateViewPlatformTransform(this.universe.getViewingPlatform().getViewPlatformTransform(),
          getViewYaw(), getViewPitch(), getViewScale());

      revalidate();
      repaint();
      if (OperatingSystem.isMacOSX()) {
        final Component root = SwingUtilities.getRoot(this);
        EventQueue.invokeLater(new Runnable() {
            public void run() {
              // Request focus again even if dialog isn't supposed to have lost focus!
              if (KeyboardFocusManager.getCurrentKeyboardFocusManager().getActiveWindow() != root) {
                root.requestFocus();
              }
            }
          });
      }
    }

    /**
     * Disposes universe bound to canvas.
     */
    private void disposeUniverse() {
      // Unlink scene to universe
      this.universe.getLocale().removeBranchGraph(this.sceneTree);
      this.universe.cleanup();
      this.universe = null;
    }

    private void updateInfo(Double elevation) {
      if (this.pickedVertex != null) {
        if (elevation == null) {
          elevation = 0d;
          String vertexElevationPropertyKey = String.format(Locale.US, Terrain3D.TERRAIN_ELEVATION_KEY_FORMAT, this.pickedVertex.x, this.pickedVertex.z);
          for (int i = 0; i < previewedRooms.size(); i++) {
            String elevationValue = previewedRooms.get(i).getProperty(vertexElevationPropertyKey);
            if (elevationValue != null) {
              try {
                elevation = Double.parseDouble(elevationValue);
                break;
              } catch (NumberFormatException ex) {
              }
            }
          }
        }
        Format formatWithUnit = preferences.getLengthUnit().getFormatWithUnit();
        infoLabel.setText(String.format(resource.getString("infoLabel.text"),
            formatWithUnit.format(this.pickedVertex.x), formatWithUnit.format(this.pickedVertex.z), formatWithUnit.format(elevation),
            magnetismEnabled ? resource.getString("activated") : resource.getString("deactivated"),
            alignmentActivated ? resource.getString("activated") : resource.getString("deactivated")));
      } else {
        infoLabel.setText(" ");
      }
    }

    /**
     * Returns the <code>yaw</code> angle used by view platform transform.
     */
    protected float getViewYaw() {
      return this.viewYaw;
    }

    /**
     * Sets the <code>yaw</code> angle used by view platform transform.
     */
    protected void setViewYaw(float viewYaw) {
      this.viewYaw = viewYaw;
      if (this.universe != null) {
        updateViewPlatformTransform(this.universe.getViewingPlatform().getViewPlatformTransform(),
            getViewYaw(), getViewPitch(), getViewScale());
      }
    }

    /**
     * Returns the zoom factor used by view platform transform.
     */
    protected float getViewScale() {
      return this.viewScale;
    }

    /**
     * Sets the zoom factor used by view platform transform.
     */
    protected void setViewScale(float viewScale) {
      this.viewScale = viewScale;
      if (this.universe != null) {
        updateViewPlatformTransform(this.universe.getViewingPlatform().getViewPlatformTransform(),
            getViewYaw(), getViewPitch(), getViewScale());
      }
    }

    /**
     * Returns the <code>pitch</code> angle used by view platform transform.
     */
    protected float getViewPitch() {
      return this.viewPitch;
    }

    /**
     * Sets the <code>pitch</code> angle used by view platform transform.
     */
    protected void setViewPitch(float viewPitch) {
      this.viewPitch = viewPitch;
      if (this.universe != null) {
        updateViewPlatformTransform(this.universe.getViewingPlatform().getViewPlatformTransform(),
            getViewYaw(), getViewPitch(), getViewScale());
      }
    }

    /**
     * Sets whether the component 3D should use parallel or perspective projection.
     */
    protected void setParallelProjection(boolean parallelProjection) {
      this.parallelProjection = parallelProjection;
      if (this.universe != null) {
        this.universe.getViewer().getView().setProjectionPolicy(parallelProjection
            ? View.PARALLEL_PROJECTION
            : View.PERSPECTIVE_PROJECTION);
      }
    }

    /**
     * Returns <code>true</code> if the component 3D uses parallel projection.
     */
    protected boolean isParallelProjection() {
      return this.parallelProjection;
    }

    /**
     * Updates the given view platform transformation from yaw angle, pitch angle and scale.
     */
    private void updateViewPlatformTransform(TransformGroup viewPlatformTransform,
                                             float viewYaw, float viewPitch,
                                             float viewScale) {
      // Default distance used to view a 2 unit wide scene
      double nominalDistanceToCenter = 2;
      // We don't use a TransformGroup in scene tree to be able to share the same scene
      // in the different views displayed by OrientationPreviewComponent class
      Transform3D translation = new Transform3D();
      translation.setTranslation(new Vector3d(0, 0, nominalDistanceToCenter));
      Transform3D pitchRotation = new Transform3D();
      pitchRotation.rotX(viewPitch);
      Transform3D yawRotation = new Transform3D();
      yawRotation.rotY(viewYaw);
      Transform3D scale = new Transform3D();
      scale.setScale(viewScale);

      pitchRotation.mul(translation);
      yawRotation.mul(pitchRotation);
      scale.mul(yawRotation);
      viewPlatformTransform.setTransform(scale);

      // Update axes transformation to show current orientation and display it in bottom left corner
      Transform3D axesTransform = new Transform3D();
      axesTransform.setScale(viewScale);
      pitchRotation.rotX(-viewPitch);
      yawRotation.rotY(-viewYaw);
      axesTransform.mul(yawRotation, axesTransform);
      axesTransform.mul(pitchRotation, axesTransform);
      translation = new Transform3D();
      translation.setTranslation(new Vector3f(-.82f * viewScale, -.82f * viewScale, .82f * viewScale));
      pitchRotation.rotX(viewPitch);
      yawRotation.rotY(viewYaw);
      axesTransform.mul(translation, axesTransform);
      axesTransform.mul(pitchRotation, axesTransform);
      axesTransform.mul(yawRotation, axesTransform);
    }

    /**
     * Returns scene tree root.
     */
    private BranchGroup createSceneTree(Home home) {
      BranchGroup root = new BranchGroup();
      root.setCapability(BranchGroup.ALLOW_DETACH);
      root.setCapability(BranchGroup.ALLOW_CHILDREN_READ);
      // Build scene tree
      root.addChild(createModelTree());
      root.addChild(createHomeNode(home));
      root.addChild(createBackgroundNode());
      root.addChild(createFeedbackNode());
      for (Light light : createLights()) {
        root.addChild(light);
      }
      return root;
    }

    /**
     * Returns the root of model tree.
     */
    private Node createModelTree() {
      TransformGroup modelTransformGroup = new TransformGroup();
      //  Allow transform group to have new children
      modelTransformGroup.setCapability(Group.ALLOW_CHILDREN_READ);
      modelTransformGroup.setCapability(Group.ALLOW_CHILDREN_WRITE);
      modelTransformGroup.setCapability(Group.ALLOW_CHILDREN_EXTEND);
      // Allow the change of the transformation that sets model size and position
      modelTransformGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      return modelTransformGroup;
    }

    /**
     * Returns the root of 3D objects of home in outline mode.
     */
    private Node createHomeNode(Home home) {
      TransformGroup homeTree = new TransformGroup();
      homeTree.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);

      Object3DBranchFactory factory = new Object3DBranchFactory(this.preferences) {
          public boolean isDrawingModeEnabled() {
            return true;
          }

          @Override
          public Object createObject3D(Home home, Selectable item, boolean waitForLoading) {
            if (item instanceof HomeFurnitureGroup) {
              Group group = new Group();
              for (HomePieceOfFurniture piece : ((HomeFurnitureGroup)item).getAllFurniture()) {
                if (!(piece instanceof HomeFurnitureGroup)) {
                  group.addChild((Node)createObject3D(home, piece, true));
                }
              }
              return group;
            } else if (item instanceof Wall
                       || item instanceof Room
                       || item instanceof HomePieceOfFurniture) {
              Node object3D = (Node)super.createObject3D(home, item, waitForLoading);
              object3D.setPickable(false);
              return object3D;
            } else {
              return null;
            }
          }
        };
      home = home.clone();
      home.getEnvironment().setDrawingMode(HomeEnvironment.DrawingMode.OUTLINE);
      for (Selectable item : home.getSelectableViewableItems()) {
        Node object = (Node)factory.createObject3D(home, item, true);
        if (object != null) {
          homeTree.addChild(object);
        }
      }
      return homeTree;
    }

    /**
     * Sets whether home items are visible in the 3D preview or not.
     */
    public void setHomeVisible(boolean visible) {
      TransformGroup homeTransformGroup = (TransformGroup)this.sceneTree.getChild(1);
      if (visible) {
        TransformGroup modelTransformGroup = (TransformGroup)this.sceneTree.getChild(0);
        Transform3D transform = new Transform3D();
        modelTransformGroup.getTransform(transform);
        homeTransformGroup.setTransform(transform);
      } else {
        Transform3D transform = new Transform3D();
        transform.setScale(0);
        homeTransformGroup.setTransform(transform);
      }
    }

    /**
     * Returns the background node.
     */
    private Node createBackgroundNode() {
      Background background = new Background(new Color3f(0.9f, 0.9f, 0.9f));
      background.setCapability(Background.ALLOW_COLOR_WRITE);
      background.setApplicationBounds(new BoundingSphere(new Point3d(0, 0, 0), 1E7));
      return background;
    }

    /**
     * Sets the background color.
     */
    public void setBackground(Color backgroundColor) {
      super.setBackground(backgroundColor);
      ((Background)this.sceneTree.getChild(2)).setColor(new Color3f(backgroundColor));
    }

    private Node createFeedbackNode() {
      TransformGroup feedbackNode = new TransformGroup();
      feedbackNode.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      feedbackNode.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
      feedbackNode.setCapability(TransformGroup.ALLOW_CHILDREN_READ);

      Transform3D scale = new Transform3D();
      scale.setScale(0);

      TransformGroup pointTransformGroup = new TransformGroup(scale);
      pointTransformGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      Appearance sphereAppearance = new Appearance();
      sphereAppearance.setColoringAttributes(new ColoringAttributes(new Color3f(0.1f, 0.1f, 0.8f), ColoringAttributes.SHADE_GOURAUD));
      sphereAppearance.setTransparencyAttributes(new TransparencyAttributes(TransparencyAttributes.NICEST, 0.2f));
      Sphere sphere = new Sphere(0.015f, 0, sphereAppearance);
      pointTransformGroup.addChild(sphere);
      feedbackNode.addChild(pointTransformGroup);

      TransformGroup influenceDiskTransformGroup = new TransformGroup(scale);
      influenceDiskTransformGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      Appearance diskAppearance = new Appearance();
      diskAppearance.setColoringAttributes(new ColoringAttributes(new Color3f(0.08f, 0.08f, 0.6f), ColoringAttributes.SHADE_GOURAUD));
      diskAppearance.setTransparencyAttributes(new TransparencyAttributes(TransparencyAttributes.BLENDED, 0.8f));
      diskAppearance.setPolygonAttributes(new PolygonAttributes(PolygonAttributes.POLYGON_FILL, PolygonAttributes.CULL_NONE, 0));
      Cylinder cylinder = new Cylinder(1f, 1f, 0, 180, 1, diskAppearance);
      Appearance cylinderAppearance = (Appearance)diskAppearance.cloneNodeComponent(true);
      cylinderAppearance.setColoringAttributes(new ColoringAttributes(new Color3f(0.2f, 0.2f, 0.9f), ColoringAttributes.SHADE_GOURAUD));
      cylinder.getShape(Cylinder.BODY).setAppearance(cylinderAppearance);
      influenceDiskTransformGroup.addChild(cylinder);
      feedbackNode.addChild(influenceDiskTransformGroup);

      TransformGroup influenceBoxTransformGroup = new TransformGroup(scale);
      influenceBoxTransformGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      Box box = new Box(0.5f, 0.5f, 0.5f, diskAppearance);
      box.getShape(Box.FRONT).setAppearance(cylinderAppearance);
      box.getShape(Box.BACK).setAppearance(cylinderAppearance);
      box.getShape(Box.LEFT).setAppearance(cylinderAppearance);
      box.getShape(Box.RIGHT).setAppearance(cylinderAppearance);
      influenceBoxTransformGroup.addChild(box);
      feedbackNode.addChild(influenceBoxTransformGroup);

      return feedbackNode;
    }

    /**
     * Sets the location of the feedback point.
     */
    public void setFeedbackLocation(float x, float y, float z, float radius, boolean box) {
      TransformGroup feedbackGroup = (TransformGroup)this.sceneTree.getChild(3);
      Transform3D pointTransformation = new Transform3D();
      Transform3D influenceDiskTransformation = new Transform3D();
      Transform3D influenceBoxTransformation = new Transform3D();
      if (Float.isNaN(x)) {
        pointTransformation.setScale(0);
        influenceDiskTransformation.setScale(0);
        influenceBoxTransformation.setScale(0);
      } else {
        Transform3D transform = new Transform3D();
        feedbackGroup.getTransform(transform);
        pointTransformation.setScale(getViewScale() / transform.getScale() * getPreferredSize().width / getWidth());
        pointTransformation.setTranslation(new Vector3f(x, y, z));
        if (box) {
          influenceBoxTransformation.setScale(new Vector3d(1E7, y + 1, 2 * radius));
          Transform3D rotation = new Transform3D();
          rotation.rotY(getViewYaw());
          rotation.setTranslation(new Vector3d(x, (y + 1) / 2, z));
          influenceBoxTransformation.mul(rotation, influenceBoxTransformation);
          influenceDiskTransformation.setScale(0);
        } else {
          influenceDiskTransformation.setScale(new Vector3d(radius, y + 1, radius));
          influenceDiskTransformation.setTranslation(new Vector3d(x, (y + 1) / 2, z));
          influenceBoxTransformation.setScale(0);
        }
      }
      ((TransformGroup)feedbackGroup.getChild(0)).setTransform(pointTransformation);
      ((TransformGroup)feedbackGroup.getChild(1)).setTransform(influenceDiskTransformation);
      ((TransformGroup)feedbackGroup.getChild(2)).setTransform(influenceBoxTransformation);
    }

    /**
     * Returns the lights of the scene.
     */
    private Light [] createLights() {
      Light [] lights = {
          new DirectionalLight(new Color3f(0.9f, 0.9f, 0.9f), new Vector3f(1.732f, -0.8f, -1)),
          new DirectionalLight(new Color3f(0.9f, 0.9f, 0.9f), new Vector3f(-1.732f, -0.8f, -1)),
          new DirectionalLight(new Color3f(0.9f, 0.9f, 0.9f), new Vector3f(0, -0.8f, 1)),
          new DirectionalLight(new Color3f(0.66f, 0.66f, 0.66f), new Vector3f(0, 1f, 0)),
          new AmbientLight(new Color3f(0.2f, 0.2f, 0.2f))};

      for (Light light : lights) {
        light.setInfluencingBounds(new BoundingSphere(new Point3d(), 1E7));
      }
      return lights;
    }

    /**
     * Sets the terrain rooms displayed by this component.
     */
    public void setTerrainRooms(final List<Room> editedRooms) {
      this.previewedRooms = editedRooms;

      TransformGroup modelTransformGroup = (TransformGroup)this.sceneTree.getChild(0);
      modelTransformGroup.removeAllChildren();
      modelTransformGroup.setTransform(new Transform3D());
      for (Room room : editedRooms) {
        modelTransformGroup.addChild(new Terrain3D(room, editedRooms));
      }
      BoundingBox bounds = ModelManager.getInstance().getBounds(modelTransformGroup);
      Point3d lower = new Point3d();
      bounds.getLower(lower);
      Point3d upper = new Point3d();
      bounds.getUpper(upper);
      Transform3D translation = new Transform3D();
      translation.setTranslation(new Vector3d(-(upper.x + lower.x) / 2, 0, -(upper.z + lower.z) / 2));
      Transform3D modelTransform = new Transform3D();
      modelTransform.setScale(1.8 / Math.max(Math.max(upper.x - lower.x, upper.z - lower.z), upper.y - lower.y));
      modelTransform.mul(translation);
      modelTransformGroup.setTransform(modelTransform);

      // Adjust size of remaining items
      ((TransformGroup)this.sceneTree.getChild(1)).setTransform(modelTransform);
      ((TransformGroup)this.sceneTree.getChild(3)).setTransform(modelTransform);
    }

    /**
     * Returns the terrain node displayed by this component.
     */
    private Group getTerrainsGroup() {
      TransformGroup modelTransformGroup = (TransformGroup)this.sceneTree.getChild(0);
      if (modelTransformGroup.numChildren() > 0) {
        return modelTransformGroup;
      } else {
        return null;
      }
    }

    /**
     * Returns a temporary content at OBJ format containing the terrain displayed by this component.
     */
    public Content exportTerrain(String terrainName) throws RecorderException {
      try {
        if (this.sceneTree.numChildren() != 0) {
          File tempZipFile = OperatingSystem.createTemporaryFile(terrainName, ".zip");
          String objFile = terrainName + ".obj";
          Group terrainsGroup = (Group)getTerrainsGroup();
          Group exportedGroup = new Group();
          // Export only terrain shapes
          for (int i = 0, n = terrainsGroup.numChildren(); i < n; i++) {
            Shape3D terrainShape = ((Terrain3D)terrainsGroup.getChild(i)).getTerrainShape();
            Shape3D exportedShape = new Shape3D();
            exportedShape.setAppearance(terrainShape.getAppearance());
            for (Enumeration<Geometry> enumGeometry = terrainShape.getAllGeometries(); enumGeometry.hasMoreElements(); ) {
              exportedShape.addGeometry(enumGeometry.nextElement());
            }
            exportedGroup.addChild(exportedShape);
          }
          OBJWriter.writeNodeInZIPFile(exportedGroup, tempZipFile, 0, objFile, "Created by RoomsTerrainGeneratorPlugin plug-in");
          return new TemporaryURLContent(new URL("jar:" + tempZipFile.toURI().toURL() + "!/" + objFile));
        } else {
          return null;
        }
      } catch (InterruptedIOException ex) {
        throw new InterruptedRecorderException("Export to OBJ interrupted");
      } catch (IOException ex) {
        throw new RecorderException("Couldn't export to OBJ", ex);
      }
    }

    /**
     * Returns the radius of the influence zone where elevation should be modified when the mouse is dragged.
     */
    public float getInfluenceRadius() {
      return this.influenceRadius;
    }

    /**
     * Sets the radius of the influence zone where elevation should be modified when the mouse is dragged.
     */
    public void setInfluenceRadius(float influenceRadius) {
      this.influenceRadius = influenceRadius;
    }
  }

  // Inspired from Room3D
  private static class Terrain3D extends Object3DBranch {
    private static final String TERRAIN_ELEVATION_KEY_PREFIX = "terrainElevationAt";
    private static final String TERRAIN_ELEVATION_KEY_FORMAT = TERRAIN_ELEVATION_KEY_PREFIX + "(%.2f,%.2f)";
    private static final Format TERRAIN_ELEVATION_FORMAT = new DecimalFormat(".###", DecimalFormatSymbols.getInstance(Locale.US));

    private static final String TERRAIN_GRID_SIZE_KEY = "terrainGridSize";
    private static final Format TERRAIN_GRID_SIZE_FORMAT = new DecimalFormat(".##", DecimalFormatSymbols.getInstance(Locale.US));

    private static final float DEFAULT_GRID_SIZE = 50f;

    private static final LineAttributes OUTLINE_LINE_ATTRIBUTES = new LineAttributes(1f, LineAttributes.PATTERN_SOLID, true);

    private List<Room> terrainRooms;

    /**
     * Creates the 3D terrain matching the given home <code>room</code>.
     */
    public Terrain3D(Room room, List<Room> terrainRooms) {
      setUserData(room);
      this.terrainRooms = new ArrayList<Room>(terrainRooms);
      // Update order to ensure that rooms are sorted first in level order
      if (room.getLevel() != null) {
        Collections.sort(this.terrainRooms, new Comparator<Room>() {
          public int compare(Room room1, Room room2) {
            int comparison = Float.compare(room1.getLevel().getElevation(), room2.getLevel().getElevation());
            if (comparison != 0) {
              return comparison;
            } else {
              return room1.getLevel().getElevationIndex() - room2.getLevel().getElevationIndex();
            }
          }
        });
      }

      // Allow room branch to be removed from its parent
      setCapability(BranchGroup.ALLOW_DETACH);
      // Allow to read branch shape children
      setCapability(BranchGroup.ALLOW_CHILDREN_READ);
      setCapability(BranchGroup.ALLOW_PICKABLE_WRITE);

      // Add room floor empty shapes to branch
      Group terrainGroup = new Group();
      terrainGroup.setCapability(Group.ALLOW_CHILDREN_READ);
      terrainGroup.addChild(createTerrainPartShape(false));
      terrainGroup.addChild(createTerrainPartShape(true));
      addChild(terrainGroup);

      // Set shape geometry and appearance
      updateTerrainGeometry();
      updateTerrainAppearance();
    }

    /**
     * Returns a new shape with no geometry
     * and a default appearance with a white material.
     */
    private Node createTerrainPartShape(boolean outline) {
      Shape3D terrainShape = new Shape3D();
      // Allow shape to change its geometry
      terrainShape.setCapability(Shape3D.ALLOW_GEOMETRY_WRITE);
      terrainShape.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
      terrainShape.setCapability(Shape3D.ALLOW_APPEARANCE_READ);

      Appearance terrainAppearance = new Appearance();
      terrainShape.setAppearance(terrainAppearance);
      terrainAppearance.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_READ);
      TransparencyAttributes transparencyAttributes = new TransparencyAttributes();
      transparencyAttributes.setCapability(TransparencyAttributes.ALLOW_VALUE_WRITE);
      transparencyAttributes.setCapability(TransparencyAttributes.ALLOW_MODE_WRITE);
      terrainAppearance.setTransparencyAttributes(transparencyAttributes);
      terrainAppearance.setCapability(Appearance.ALLOW_RENDERING_ATTRIBUTES_READ);
      RenderingAttributes renderingAttributes = new RenderingAttributes();
      renderingAttributes.setCapability(RenderingAttributes.ALLOW_VISIBLE_WRITE);
      terrainAppearance.setRenderingAttributes(renderingAttributes);

      if (outline) {
        terrainAppearance.setColoringAttributes(Object3DBranch.OUTLINE_COLORING_ATTRIBUTES);
        terrainAppearance.setPolygonAttributes(Object3DBranch.OUTLINE_POLYGON_ATTRIBUTES);
        terrainAppearance.setLineAttributes(Object3DBranch.OUTLINE_LINE_ATTRIBUTES);
      } else {
        terrainAppearance.setCapability(Appearance.ALLOW_MATERIAL_WRITE);
        terrainAppearance.setMaterial(DEFAULT_MATERIAL);
        terrainAppearance.setCapability(Appearance.ALLOW_TEXTURE_WRITE);
        terrainAppearance.setCapability(Appearance.ALLOW_TEXTURE_READ);
        terrainAppearance.setCapability(Appearance.ALLOW_TEXTURE_ATTRIBUTES_WRITE);
      }

      return terrainShape;
    }

    @Override
    public void update() {
      updateTerrainGeometry();
      updateTerrainAppearance();
    }

    public Shape3D getTerrainShape() {
      return (Shape3D)((Group)getChild(0)).getChild(0);
    }

    /**
     * Sets the 3D geometry of the shapes that matches room 2D geometry.
     */
    private void updateTerrainGeometry() {
      Group terrainGroup = (Group)getChild(0);
      Shape3D terrainFilledShape = (Shape3D)terrainGroup.getChild(0);
      Shape3D terrainOutlineShape = terrainGroup.numChildren() > 1
          ? (Shape3D)terrainGroup.getChild(1)
          : null;
      int currentGeometriesCount = terrainFilledShape.numGeometries();
      Room room = (Room)getUserData();
      if (room.getLevel() == null || room.getLevel().isViewableAndVisible()) {
        for (Geometry terrainGeometry : createTerrainGeometries(((Room)getUserData()).getFloorTexture())) {
          terrainFilledShape.addGeometry(terrainGeometry);
          if (terrainOutlineShape != null) {
            terrainOutlineShape.addGeometry(terrainGeometry);
          }
        }
      }
      for (int i = currentGeometriesCount - 1; i >= 0; i--) {
        terrainFilledShape.removeGeometry(i);
        if (terrainOutlineShape != null) {
          terrainOutlineShape.removeGeometry(i);
        }
      }
      setPickable(true);
    }

    /**
     * Returns terrain geometry computed from its points.
     */
    private Geometry [] createTerrainGeometries(HomeTexture texture) {
      Room room = (Room)getUserData();
      float [][] points = room.getPoints();
      if (points.length > 2) {
        // Retrieve room points
        List<float [][]> terrainPoints;
        List<float [][]> terrainHoles;
        List<float [][]> terrainPointsWithoutHoles;
        Area terrainVisibleArea;
        // If room isn't singular retrieve all the points of its different polygons
        if (!room.isSingular()
            || this.terrainRooms.get(this.terrainRooms.size() - 1) != room) {
          terrainVisibleArea = new Area(getShape(points));
          // Remove other rooms surface that may overlap the current room
          for (int i = this.terrainRooms.size() - 1; i > 0 && this.terrainRooms.get(i) != room; i--) {
            Room otherRoom = this.terrainRooms.get(i);
            terrainVisibleArea.subtract(new Area(getShape(otherRoom.getPoints())));
          }
          terrainPoints = new ArrayList<float[][]>();
          terrainHoles = new ArrayList<float[][]>();
          terrainPointsWithoutHoles = getAreaPoints(terrainVisibleArea, terrainPoints, terrainHoles, 1, false);
        } else {
          if (room.isClockwise()) {
            // Reverse points order
            points = getReversedArray(points);
          }
          terrainPointsWithoutHoles =
          terrainPoints = Arrays.asList(new float [][][] {points});
          terrainHoles = Collections.emptyList();
          terrainVisibleArea = null;
        }

        List<Geometry> geometries = new ArrayList<Geometry> (3);
        float gridSize = DEFAULT_GRID_SIZE;
        for (Room terrainRoom : this.terrainRooms) {
          try {
            gridSize = Float.parseFloat(terrainRoom.getProperty(TERRAIN_GRID_SIZE_KEY));
            break;
          } catch (NumberFormatException ex) {
          } catch (NullPointerException ex) {
          }
        }
        if (!terrainPointsWithoutHoles.isEmpty()) {
          // Compute room geometry
          List<float [][]> geometryPointsWithoutHoles = new ArrayList<float[][]>();
          List<Area> terrainPartAreas = new ArrayList<Area>();
          for (int i = 0; i < terrainPointsWithoutHoles.size(); i++) {
            float [][] terrainPartPoints = terrainPointsWithoutHoles.get(i);
            // Subdivide area in smaller squares to ensure a smoother effect
            float xMin = Float.MAX_VALUE;
            float xMax = Float.MIN_VALUE;
            float zMin = Float.MAX_VALUE;
            float zMax = Float.MIN_VALUE;
            for (float [] point : terrainPartPoints) {
              xMin = Math.min(xMin, point [0]);
              xMax = Math.max(xMax, point [0]);
              zMin = Math.min(zMin, point [1]);
              zMax = Math.max(zMax, point [1]);
            }
            xMin -= xMin < 0 ? xMin % gridSize + gridSize : xMin % gridSize;
            zMin -= zMin < 0 ? zMin % gridSize + gridSize : zMin % gridSize;

            Area terrainPartArea = new Area(getShape(terrainPartPoints));
            terrainPartAreas.add(terrainPartArea);
            for (float xSquare = xMin; xSquare < xMax; xSquare += gridSize) {
              for (float zSquare = zMin; zSquare < zMax; zSquare += gridSize) {
                Area terrainPartSquare = new Area(new Rectangle2D.Float(xSquare, zSquare, gridSize, gridSize));
                terrainPartSquare.intersect(terrainPartArea);
                if (!terrainPartSquare.isEmpty()) {
                  geometryPointsWithoutHoles.addAll(getAreaPoints(terrainPartSquare, 1, false));
                }
              }
            }
          }
          List<float []> pointElevations = new ArrayList<float[]>(geometryPointsWithoutHoles.size());
          for (int i = 0; i < geometryPointsWithoutHoles.size(); i++) {
            float [][] terrainPartAreaPoints = geometryPointsWithoutHoles.get(i);
            float [] terrainPartAreaPointElevations = new float [terrainPartAreaPoints.length];
            for (int k = 0; k < terrainPartAreaPoints.length; k++) {
              terrainPartAreaPointElevations [k] = 0;
              String vertexElevationPropertyKey = String.format(Locale.US, TERRAIN_ELEVATION_KEY_FORMAT, terrainPartAreaPoints [k][0], terrainPartAreaPoints [k][1]);
              String elevation = room.getProperty(vertexElevationPropertyKey);
              if (elevation != null) {
                try {
                  terrainPartAreaPointElevations [k] = Float.parseFloat(elevation);
                } catch (NumberFormatException ex) {
                }
              }
            }
            pointElevations.add(terrainPartAreaPointElevations);
          }

          geometries.add(computeTerrainPartGeometry(geometryPointsWithoutHoles, pointElevations, texture));

          // Compute border geometry
          Geometry borderGeometry = computeTerrainBorderGeometry(geometryPointsWithoutHoles, pointElevations, texture, terrainPartAreas);
          if (borderGeometry != null) {
            geometries.add(borderGeometry);
          }
        }


        return geometries.toArray(new Geometry [geometries.size()]);
      } else {
        return new Geometry [0];
      }
    }

    /**
     * Returns the terrain part geometry matching the given points.
     */
    private Geometry computeTerrainPartGeometry(List<float [][]> geometryPoints,
                                                List<float []> terrainPointElevations,
                                                HomeTexture texture) {
      int [] stripCounts = new int [geometryPoints.size()];
      int vertexCount = 0;
      for (int i = 0; i < geometryPoints.size(); i++) {
        float [][] areaPoints = geometryPoints.get(i);
        stripCounts [i] = areaPoints.length;
        vertexCount += stripCounts [i];
      }
      Point3f [] coords = new Point3f [vertexCount];
      int i = 0;
      for (int j = 0; j < geometryPoints.size(); j++) {
        float [][] areaPoints = geometryPoints.get(j);
        float [] terrainPartPointElevations = terrainPointElevations.get(j);
        for (int k = 0; k < areaPoints.length; k++) {
          float y = terrainPartPointElevations [k];
          coords [i++] = new Point3f(areaPoints [k][0], y, areaPoints [k][1]);
        }
      }
      GeometryInfo geometryInfo = new GeometryInfo(GeometryInfo.POLYGON_ARRAY);
      geometryInfo.setCoordinates(coords);
      geometryInfo.setStripCounts(stripCounts);

      if (texture != null) {
        TexCoord2f [] textureCoords = new TexCoord2f [vertexCount];
        i = 0;
        // Compute texture coordinates
        for (float [][] areaPoints : geometryPoints) {
          for (int k = 0; k < areaPoints.length; k++) {
            textureCoords [i++] = new TexCoord2f(areaPoints [k][0], -areaPoints [k][1]);
          }
        }
        geometryInfo.setTextureCoordinateParams(1, 2);
        geometryInfo.setTextureCoordinates(0, textureCoords);
      }

      // Generate normals
      new NormalGenerator().generateNormals(geometryInfo);
      return geometryInfo.getIndexedGeometryArray();
    }

    /**
     * Returns the terrain border geometry matching the geometry points which belong to terrain areas.
     */
    private Geometry computeTerrainBorderGeometry(List<float [][]> geometryPoints,
                                                  List<float []>   pointElevations,
                                                  HomeTexture texture,
                                                  List<Area> terrainPartAreas) {
      List<Point3f> coords = new ArrayList<Point3f>();
      List<TexCoord2f> textureCoords = texture != null
          ? new ArrayList<TexCoord2f>()
          : null;
      for (int i = 0; i < geometryPoints.size(); i++) {
        float [][] areaPoints = geometryPoints.get(i);
        float [] elevations = pointElevations.get(i);
        for (int j = 0; j < areaPoints.length; j++) {
          boolean border = false;
          int nextPoint = j < areaPoints.length - 1
              ? j + 1
              : 0;
          for (Area shape : terrainPartAreas) {
            if (!shape.contains(areaPoints [j][0] - 1E-3, areaPoints [j][1] - 1E-3, 2E-3, 2E-3)
                && !shape.contains(areaPoints [nextPoint][0] - 1E-3, areaPoints [nextPoint][1] - 1E-3, 2E-3, 2E-3)) {
              border = true;
              break;
            }
          }
          if (border
              && (elevations [j] > 0 || elevations [nextPoint] > 0)) {
            coords.add(new Point3f(areaPoints [j][0], elevations [j], areaPoints [j][1]));
            coords.add(new Point3f(areaPoints [j][0], 0, areaPoints [j][1]));
            coords.add(new Point3f(areaPoints [nextPoint][0], 0, areaPoints [nextPoint][1]));
            coords.add(new Point3f(areaPoints [nextPoint][0], elevations [nextPoint], areaPoints [nextPoint][1]));
            if (textureCoords != null) {
              textureCoords.add(new TexCoord2f(0, elevations [j]));
              textureCoords.add(new TexCoord2f(0, 0));
              float textureCoord = (float)(Point2D.distance(areaPoints [j][0], areaPoints [j][1],
                  areaPoints [nextPoint][0], areaPoints [nextPoint][1]));
              textureCoords.add(new TexCoord2f(textureCoord, 0));
              textureCoords.add(new TexCoord2f(textureCoord, elevations [nextPoint]));
            }
          }
        }
      }

      if (coords.size() > 0) {
        GeometryInfo geometryInfo = new GeometryInfo(GeometryInfo.QUAD_ARRAY);
        geometryInfo.setCoordinates(coords.toArray(new Point3f [coords.size()]));
        if (textureCoords != null) {
          geometryInfo.setTextureCoordinateParams(1, 2);
          geometryInfo.setTextureCoordinates(0, textureCoords.toArray(new TexCoord2f [textureCoords.size()]));
        }

        // Generate normals
        new NormalGenerator(Math.PI / 8).generateNormals(geometryInfo);
        return geometryInfo.getIndexedGeometryArray();
      } else {
        return null;
      }
    }

    /**
     * Returns an array that cites <code>points</code> in reverse order.
     */
    private float [][] getReversedArray(float [][] points) {
      points = points.clone();
      List<float []> pointList = Arrays.asList(points);
      Collections.reverse(pointList);
      return pointList.toArray(points);
    }

    /**
     * Sets terrain appearance with its color, texture.
     */
    private void updateTerrainAppearance() {
      Room room = (Room)getUserData();
      Group roomFloorGroup = (Group)getChild(0);
      updateFilledTerrainAppearance(((Shape3D)roomFloorGroup.getChild(0)).getAppearance(),
          room.getFloorTexture(), room.getFloorColor(), room.getFloorShininess());
    }

    /**
     * Sets filled terrain appearance with its color, texture and visibility.
     */
    private void updateFilledTerrainAppearance(final Appearance terrainAppearance,
                                               final HomeTexture terrainTexture,
                                               Integer terrainColor,
                                               float shininess) {
      if (terrainTexture == null) {
        terrainAppearance.setMaterial(getMaterial(terrainColor, terrainColor, shininess));
        terrainAppearance.setTexture(null);
      } else {
        // Update material and texture of terrain
        terrainAppearance.setMaterial(getMaterial(DEFAULT_COLOR, DEFAULT_AMBIENT_COLOR, shininess));
        terrainAppearance.setTextureAttributes(getTextureAttributes(terrainTexture, true));
        final TextureManager textureManager = TextureManager.getInstance();
        textureManager.loadTexture(terrainTexture.getImage(), false,
            new TextureManager.TextureObserver() {
                public void textureUpdated(Texture texture) {
                  texture = getHomeTextureClone(texture, null);
                  if (terrainAppearance.getTexture() != texture) {
                    terrainAppearance.setTexture(texture);
                  }
                }
              });
      }
    }
  }
}
